/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client;

import java.io.IOException;
import java.net.InetAddress;
import java.util.Collections;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;

import net.yale.cs.p4p.api.data.InetService;
import net.yale.cs.p4p.p2p.azureus.client.data.PDistCache;
import net.yale.cs.p4p.p2p.azureus.client.data.PIDMapCache;
import net.yale.cs.p4p.p2p.azureus.client.data.PeerCache;
import net.yale.cs.p4p.p2p.azureus.client.data.Torrents;
import net.yale.cs.p4p.p2p.azureus.client.data.selection.PeerConnectionMonitor;
import net.yale.cs.p4p.p2p.azureus.client.data.selection.PeerSelectionP4PSimple;
import net.yale.cs.p4p.p2p.azureus.client.experiment.ExperimentDecider;
import net.yale.cs.p4p.p2p.azureus.client.experiment.ExperimentQuery;
import net.yale.cs.p4p.p2p.azureus.client.listeners.P4PDownloadPeerListener;
import net.yale.cs.p4p.p2p.azureus.client.listeners.P4PDownloadTrackerListener;
import net.yale.cs.p4p.p2p.azureus.client.ui.MainView;
import net.yale.cs.p4p.portal.discovery.DiscoveryInterface;
import net.yale.cs.p4p.portal.discovery.DiscoveryPortalDB;

import org.gudy.azureus2.plugins.PluginException;
import org.gudy.azureus2.plugins.PluginInterface;
import org.gudy.azureus2.plugins.PluginListener;
import org.gudy.azureus2.plugins.UnloadablePlugin;
import org.gudy.azureus2.plugins.ipc.IPCException;
import org.gudy.azureus2.plugins.ipc.IPCInterface;
import org.gudy.azureus2.plugins.logging.LoggerChannel;

public class P4PPlugin implements UnloadablePlugin
{
	public static final String PLUGIN_NAME = "P4P";

	public static final boolean IS_DEBUG = false;
	public static final boolean USE_UI = false;

	private class P4PPluginListener implements PluginListener {
		public void closedownComplete() {
			finishUnload();
		}

		public void closedownInitiated() {
			startUnload();
			finishUnload();
		}
		public void initializationComplete() {
			Timer timer = new Timer();
			timer.schedule(new TimerTask() {
					public void run() {
						doUpdateInstrumentationComponents();
					}
					},
					5 * 1000);
		}
	}

	private static final Map<String, String> SELECTION_PARAMS = Collections.unmodifiableMap(new TreeMap<String, String>());

	private Config config_;
	public Config getConfig() { return config_; }

	private PluginInterface pluginInterface_;
	public PluginInterface getPluginInterface() { return pluginInterface_; }

	private LoggerChannel logger_;
	public LoggerChannel getLogger() { return logger_; }

	private Torrents torrents_;
	public Torrents getTorrents() { return torrents_; }

	private PeerCache peerCache_;
	public PeerCache getPeerCache() { return peerCache_; }

	private PIDMapCache pidMapCache_;
	public PIDMapCache getPIDMapCache() { return pidMapCache_; }

	private PDistCache pdistCache_;
	public PDistCache getPDistCache() { return pdistCache_; }

	private PeerConnectionMonitor peerMonitor_;
	public PeerConnectionMonitor getPeerMonitor() { return peerMonitor_; }
	
	private ExperimentDecider experiment_;
	public ExperimentDecider getExperiment() { return experiment_; }

	private P4PPluginListener pluginListener_;
	private P4PDownloadPeerListener downloadPeerListener_;
	private P4PDownloadTrackerListener downloadTrackerListener_;

	private MainView ui_;

	public void initialize(PluginInterface pluginInterface) throws PluginException {
		pluginInterface_ = pluginInterface;

		/* Create a logger for our plugin */
		logger_ = pluginInterface_.getLogger().getChannel(PLUGIN_NAME);

		/* Create configuration options */
		config_ = new Config(this);
		
		torrents_ = new Torrents(this);

		/* Start background threads */
		getLogger().log("Starting background threads");
		experiment_ = new ExperimentDecider(this);
		experiment_.start();
		peerCache_ = new PeerCache(this);
		peerCache_.start();
		pidMapCache_ = new PIDMapCache(this,
				getConfig().getPIDMapQueryInterval() * 1000,
				peerCache_);
		pidMapCache_.start();
		pdistCache_ = new PDistCache(this,
				getConfig().getPDistQueryInterval() * 1000,
				peerCache_);
		pdistCache_.start();
		peerMonitor_ = new PeerConnectionMonitor(this, new PeerSelectionP4PSimple(this));

		/* Register listeners */
		getLogger().log("Installing listeners");
		pluginListener_ = new P4PPluginListener();
		getPluginInterface().addListener(pluginListener_);
		downloadPeerListener_ = new P4PDownloadPeerListener(this);
		getPluginInterface().getDownloadManager().getGlobalDownloadEventNotifier().addPeerListener(downloadPeerListener_);
		downloadTrackerListener_ = new P4PDownloadTrackerListener(this);
		getPluginInterface().getDownloadManager().getGlobalDownloadEventNotifier().addTrackerListener(downloadTrackerListener_);
		
		/* Register UI */
		if (USE_UI) {
			ui_ = new MainView(this);
		}

		getLogger().log("Plugin loaded");
	}

	private void startUnload() {
		if (logger_ != null)
			logger_.log("Stopping background threads");
		if (experiment_ != null)
			experiment_.interrupt();
		if (pdistCache_ != null)
			pdistCache_.interrupt();
		if (pidMapCache_ != null)
			pidMapCache_.interrupt();
		if (peerCache_ != null)
			peerCache_.interrupt();
	}

	private void finishUnload() {
		if (USE_UI) {
			if (logger_ != null)
				logger_.log("Removing UI");
			if (ui_ != null) {
				ui_.cleanup();
				ui_ = null;
			}
		}
		
		if (logger_ != null)
			logger_.log("Removing listeners");
		
		if (downloadTrackerListener_ != null) {
			downloadTrackerListener_.cleanup();
			pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().removeTrackerListener(downloadTrackerListener_);
			downloadTrackerListener_ = null;
		}
		if (downloadPeerListener_ != null) {
			downloadPeerListener_.cleanup();
			pluginInterface_.getDownloadManager().getGlobalDownloadEventNotifier().removePeerListener(downloadPeerListener_);
			downloadPeerListener_ = null;
		}
		if (pluginListener_ != null) {
			pluginInterface_.removeListener(pluginListener_);
			pluginListener_ = null;
		}
		
		if (logger_ != null)
			logger_.log("Waiting for background threads");
		if (experiment_ != null) {
			experiment_.joinNoThrow();
			experiment_ = null;
		}
		if (peerMonitor_ != null)
			peerMonitor_ = null;
		if (pdistCache_ != null) {
			pdistCache_.joinNoThrow();
			pdistCache_ = null;
		}
		if (pidMapCache_ != null) {
			pidMapCache_.joinNoThrow();
			pidMapCache_ = null;
		}
		if (peerCache_ != null) {
			peerCache_.joinNoThrow();
			peerCache_ = null;
		}
		
		if (torrents_ == null) {
			torrents_ = null;
		}
		
		if (config_ != null) {
			config_.cleanup();
			config_ = null;
		}

		if (logger_ != null) {
			logger_.log("Plugin unloaded");
			logger_ = null;
		}
	}

	public void unload() throws PluginException {
		startUnload();
		finishUnload();
	}

	public DiscoveryInterface getPortalDiscoveryInterface() throws IOException {
		return new DiscoveryPortalDB(
				new InetService(InetAddress.getByName(getConfig().getPortalDBHost()),
						getConfig().getPortalDBPort()));
	}
	
	public ExperimentQuery getExperimentInterface() throws IOException {
		return new ExperimentQuery(
				new InetService(InetAddress.getByName(getConfig().getExperimentHost()),
						getConfig().getExperimentPort()));
	}

	/* IPC method called by instrumentation plugin */
	public Map<String, String> getInstrumentationParams() {
		return SELECTION_PARAMS;
	}
	
	private void doUpdateInstrumentationComponents() {
		PluginInterface myPlugin = pluginInterface_;
		if (myPlugin == null)
			return;

		/* Find instrumentation plugin if available */
		PluginInterface plugin = myPlugin.getPluginManager().getPluginInterfaceByID("yaleinstrumentation");
		if (plugin == null)
			return;

		/* Get IPC interface to the plugin */
		IPCInterface ipc = plugin.getIPC();
		if (ipc == null)
			return;

		/* call the update components */
		try {
			ipc.invoke("updateComponents", new Object[] {});
		} catch (IPCException e) {
			return;
		}
	}
	
}
