/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.data;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import net.yale.cs.p4p.api.APIBase.HTTPErrorException;
import net.yale.cs.p4p.api.data.InetPrefix;
import net.yale.cs.p4p.api.data.InetService;
import net.yale.cs.p4p.api.data.PID;
import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.util.AddrUtils;
import net.yale.cs.p4p.portal.api.QueryREST;
import se.kth.nada.router.Entry;
import se.kth.nada.router.RoutingTable;

public final class PIDMapCache extends QueryCache
{

	private static final String PIDMAPFILE_BASENAME = "PIDMap-";

	private final class PIDMapQuery extends QueryTask {

		public PIDMapQuery(int interval) {
			super(interval);
		}

		private File makePIDMapFilename(long timestamp) {
			return new File(getPlugin().getPluginInterface().getPluginDirectoryName(), PIDMAPFILE_BASENAME + timestamp);
		}

		private File getPIDMapFile() {
			/* Find the latest one */
			File result = null;
			File baseDir = new File(getPlugin().getPluginInterface().getPluginDirectoryName());
			for (File f : baseDir.listFiles()) {
				if (!f.getName().startsWith(PIDMAPFILE_BASENAME))
					continue;
				if (result == null || getPIDMapTimestamp(f) > getPIDMapTimestamp(result))
					result = f;
			}
			return result;
		}

		private long getPIDMapTimestamp(File file) {

			String name = file.getName();
			if (!name.startsWith(PIDMAPFILE_BASENAME))
				return 0;
			
			int sep = name.indexOf('-');
			if (sep < 0)
				return 0;

			try {
				return Long.valueOf(name.substring(sep + 1));
			} catch (NumberFormatException e) {
				return 0;
			}
		}

		private void loadFromDisk() {
			try {
				File file = getPIDMapFile();
				if (file == null)
					return;

				getPlugin().getLogger().log("Loading PID map from file " + file);

				InputStream stream = new GZIPInputStream(new FileInputStream(file));
				try {
					pidMap_ = new RoutingTable(stream);
					lastQuery_ = getPIDMapTimestamp(file);
				} finally {
					stream.close();
				}
			} catch (IOException e) {
				getPlugin().getLogger().log("Failed to load PID Map from disk: " + e);
			}
		}

		private void saveToDisk() {
			if (pidMap_ == null)
				return;

			try {
				File existingFile = getPIDMapFile();
				try {
					File newFile = makePIDMapFilename(lastQuery_);
					getPlugin().getLogger().log("Saving PID map to file " + newFile);
					OutputStream stream = new GZIPOutputStream(new FileOutputStream(newFile));
					try {
						pidMap_.save(stream);
					} finally {
						stream.close();
					}
				} finally {
					if (existingFile != null)
						existingFile.delete();
				}
			} catch (IOException e) {
				getPlugin().getLogger().log("Failed to save PID Map to disk: " + e);
			}
		}

		protected void query() throws InterruptedException {
			
			if (needDiskLoad())
				loadFromDisk();
			
			/* Ignore if no portal server available */
			InetService portalServer = peerCache_.getPortalServer();
			if (portalServer == null)
				return;

			if (pidMap_ != null && !needPIDMapQuery())
				return;

			Map<PID, Collection<InetPrefix>> pidMap;

			/* Query and replace the existing PID Map */
			getPlugin().getLogger().log("Querying updated PID Map");
			try {
				pidMap = new QueryREST(portalServer).getPIDMap();
				getPlugin().getLogger().log("Finished querying for updated PID Map");
			} catch (HTTPErrorException e) {
				getPlugin().getLogger().log("Portal discovery failed: HTTP error " + e.getCode() + " " + e.getReason());
				return;
			} catch (IOException e) {
				getPlugin().getLogger().log("Failed to update PID Map from Portal Server: " + e);
				return;
			}


			/* Construct new PID map */
			getPlugin().getLogger().log("Constructing new PID Map");
			ArrayList<Entry> tableEntries = new ArrayList<Entry>();
			for (Map.Entry<PID, Collection<InetPrefix>> e : pidMap.entrySet()) {
				for (InetPrefix prefix : e.getValue()) {
					int addr = AddrUtils.addressToInt(prefix.getIP());
					if (addr == AddrUtils.INVALID_ADDRESS)
						continue;
					tableEntries.add(new Entry(addr, prefix.getLength(), e.getKey()));
				}
			}

			/* Use new map */
			getPlugin().getLogger().log("Applying new PID Map");
			pidMap_ = new RoutingTable(tableEntries);
			lastQuery_ = System.currentTimeMillis();
			
			saveToDisk();
		}
	}

	private PeerCache peerCache_;
	
	private boolean needDiskLoad_ = true;
	private synchronized boolean needDiskLoad() {
		boolean result = needDiskLoad_;
		needDiskLoad_ = false;
		return result;
	}

	private RoutingTable pidMap_ = null;
	public RoutingTable getPIDMap() { return pidMap_; }

	/* Time of last query for PID Map */
	private long lastQuery_ = 0;
	private synchronized boolean needPIDMapQuery() {
		return lastQuery_ + getPlugin().getConfig().getPIDMapQueryInterval() * 1000 < System.currentTimeMillis();
	}

	public PIDMapCache(P4PPlugin plugin, int queryInterval, PeerCache peerCache) {
		super(plugin, 1);
		peerCache_ = peerCache;
	}

	protected QueryTask makeQueryTask() {
		return new PIDMapQuery(5 * 1000);
	}

}
